# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dddp', '0005_auto_20150427_1209'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='connection',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='connection',
            name='session',
        ),
    ]



"""Django DDP utils for DDP messaging."""
from copy import deepcopy
from dddp import THREAD_LOCAL as this
from django.db.models.expressions import ExpressionNode


def obj_change_as_msg(obj, msg):
    """Generate a DDP msg for obj with specified msg type."""
    # check for F expressions
    exps = [
        name for name, val in vars(obj).items()
        if isinstance(val, ExpressionNode)
    ]
    if exps:
        # clone and update obj with values but only for the expression fields
        obj = deepcopy(obj)
        # Django 1.8 makes obj._meta public --> pylint: disable=W0212
        for name, val in obj._meta.model.objects.values(*exps).get(pk=obj.pk):
            setattr(obj, name, val)

    # run serialization now that all fields are "concrete" (not F expressions)
    serializer = this.serializer
    data = serializer.serialize([obj])[0]

    # collection name is <app>.<model>
    name = data['model']

    # cast ID as string
    if not isinstance(data['pk'], basestring):
        data['pk'] = '%d' % data['pk']

    payload = {
        'msg': msg,
        'collection': name,
        'id': data['pk'],
    }
    if msg != 'removed':
        payload['fields'] = data['fields']

    return (name, payload)



"""
A Python "serializer". Doesn't do much serializing per se -- just converts to
and from basic Python data types (lists, dicts, strings, etc.). Useful as a basis for
other serializers.
"""
from __future__ import unicode_literals

from django.apps import apps
from django.conf import settings
from django.core.serializers import base
from django.core.serializers import python
from django.db import DEFAULT_DB_ALIAS, models
from django.utils import six
from django.utils.encoding import force_text, is_protected_type
from dddp.models import get_meteor_id, get_object_id


class Serializer(python.Serializer):
    """
    Serializes a QuerySet to basic Python objects.
    """

    def get_dump_object(self, obj):
        data = super(Serializer, self).get_dump_object(obj)
        data["pk"] = get_meteor_id(obj)
        return data

    def handle_fk_field(self, obj, field):
        value = getattr(obj, field.name)
        self._current[field.column] = get_meteor_id(value)

    def handle_m2m_field(self, obj, field):
        if field.rel.through._meta.auto_created:
            m2m_value = lambda value: get_meteor_id(value)
            self._current['%s_ids' % field.name] = [m2m_value(related)
                               for related in getattr(obj, field.name).iterator()]


def Deserializer(object_list, **options):
    """
    Deserialize simple Python objects back into Django ORM instances.

    It's expected that you pass the Python objects themselves (instead of a
    stream or a string) to the constructor
    """
    db = options.pop('using', DEFAULT_DB_ALIAS)
    ignore = options.pop('ignorenonexistent', False)

    for d in object_list:
        # Look up the model and starting build a dict of data for it.
        try:
            Model = _get_model(d["model"])
        except base.DeserializationError:
            if ignore:
                continue
            else:
                raise
        data = {}
        if 'pk' in d:
            data[Model._meta.pk.attname] = Model._meta.pk.to_python(
                get_object_id(Model, d.get("pk", None)),
            )
        m2m_data = {}
        field_names = {f.name for f in Model._meta.fields}
        field_name_map = {
            f.column: f.name
            for f in Model._meta.fields
        }
        for field in Model._meta.many_to_many:
            field_name_map.setdefault('%s_ids' % field.name, field.name)

        # Handle each field
        for (field_column, field_value) in six.iteritems(d["fields"]):
            field_name = field_name_map.get(field_column, None)

            if ignore and field_name not in field_names:
                # skip fields no longer on model
                continue

            if isinstance(field_value, str):
                field_value = force_text(
                    field_value, options.get("encoding", settings.DEFAULT_CHARSET), strings_only=True
                )

            field = Model._meta.get_field(field_name)

            # Handle M2M relations
            if field.rel and isinstance(field.rel, models.ManyToManyRel):
                m2m_data[field.name] = [get_object_id(field.rel.to, pk) for pk in field_value]

            # Handle FK fields
            elif field.rel and isinstance(field.rel, models.ManyToOneRel):
                if field_value is not None:
                    field_value= get_object_id(field.rel.to, field_value)
                    data[field.attname] = field.rel.to._meta.get_field(field.rel.field_name).to_python(field_value)
                else:
                    data[field.attname] = None

            # Handle all other fields
            else:
                data[field.name] = field.to_python(field_value)

        obj = base.build_instance(Model, data, db)
        yield base.DeserializedObject(obj, m2m_data)


def _get_model(model_identifier):
    """
    Helper to look up a model from an "app_label.model_name" string.
    """
    try:
        return apps.get_model(model_identifier)
    except (LookupError, TypeError):
        raise base.DeserializationError("Invalid model identifier: '%s'" % model_identifier)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dddp', '0007_auto_20150505_1302'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='subscription',
            name='publication_class',
        ),
    ]



"""Django DDP API, Collections, Cursors and Publications."""
from __future__ import absolute_import, unicode_literals

# standard library
import collections
from copy import deepcopy
import traceback

# requirements
import dbarray
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import connection, connections
from django.db.models import aggregates, Q
from django.db.models.expressions import ExpressionNode
from django.db.models.sql import aggregates as sql_aggregates
from django.utils.encoding import force_text
from django.db import DatabaseError
from django.db.models import signals
import ejson

# django-ddp
from dddp import AlreadyRegistered, THREAD_LOCAL as this
from dddp.models import Connection, Subscription, get_meteor_id


XMIN = {'select': {'xmin': "'xmin'"}}


class Sql(object):

    """Extensions to django.db.models.sql.aggregates module."""

    class Array(sql_aggregates.Aggregate):

        """Array SQL aggregate extension."""

        lookup_name = 'array'
        sql_function = 'array_agg'

sql_aggregates.Array = Sql.Array


class Array(aggregates.Aggregate):

    """Array aggregate function."""

    func = 'ARRAY'
    name = 'Array'

    def add_to_query(self, query, alias, col, source, is_summary):
        """Override source field internal type so the raw array is returned."""
        class ArrayField(dbarray.ArrayFieldBase, source.__class__):

            """ArrayField for override."""

            __metaclass__ = dbarray.ArrayFieldMetaclass

            @staticmethod
            def get_internal_type():
                """Return ficticious type so Django doesn't cast as int."""
                return 'ArrayType'

        new_source = ArrayField()
        super(Array, self).add_to_query(
            query, alias, col, new_source, is_summary,
        )


def api_endpoint(path_or_func):
    """Decorator to mark a method as an API endpoint for later registration."""
    if callable(path_or_func):
        path_or_func.api_path = path_or_func.__name__
        return path_or_func
    else:
        def _api_endpoint(func):
            """Decorator inner."""
            func.api_path = path_or_func
            return func
        return _api_endpoint


def api_endpoints(obj):
    """Iterator over all API endpoint names and callbacks."""
    for name in dir(obj):
        attr = getattr(obj, name)
        api_path = getattr(attr, 'api_path', None)
        if api_path:
            yield (
                '%s%s' % (obj.api_path_prefix, api_path),
                attr,
            )
    for api_provider in obj.api_providers:
        for api_path, attr in api_endpoints(api_provider):
            yield (api_path, attr)


class APIMeta(type):

    """DDP API metaclass."""

    def __new__(mcs, name, bases, attrs):
        """Create a new APIMixin class."""
        attrs['name'] = attrs.pop('name', None) or name
        name_format = attrs.get('name_format', None)
        if name_format:
            attrs['name'] = name_format.format(**attrs)
        api_path_prefix_format = attrs.get('api_path_prefix_format', None)
        if attrs.get('api_path_prefix', None) is not None:
            pass
        elif api_path_prefix_format is not None:
            attrs['api_path_prefix'] = api_path_prefix_format.format(**attrs)
        return super(APIMeta, mcs).__new__(mcs, name, bases, attrs)


class APIMixin(object):

    """Mixin to support finding API endpoints for class instances."""

    api_path_prefix_format = None

    api_providers = []
    api_path_prefix = '/'
    _api_path_cache = None

    def api_path_map(self):
        """Cached dict of api_path: func."""
        if self._api_path_cache is None:
            self._api_path_cache = {
                api_path: func
                for api_path, func
                in api_endpoints(self)
            }
        return self._api_path_cache

    def clear_api_path_map_cache(self):
        """Clear out cache for api_path_map."""
        self._api_path_cache = None
        for api_provider in self.api_providers:
            if api_provider.clear_api_path_map_cache.im_self is not None:
                api_provider.clear_api_path_map_cache()

    def api_endpoint(self, api_path):
        """Return API endpoint for given api_path."""
        return self.api_path_map()[api_path]


def model_name(model):
    """Return model name given model class."""
    # Django supports model._meta -> pylint: disable=W0212
    return force_text(model._meta)


COLLECTION_PATH_FORMAT = '/{name}/'


class CollectionMeta(APIMeta):

    """DDP Collection metaclass."""

    def __new__(mcs, name, bases, attrs):
        """Create a new Collection class."""
        attrs.update(
            api_path_prefix_format=COLLECTION_PATH_FORMAT,
        )
        model = attrs.get('model', None)
        if attrs.get('name', None) is None and model is not None:
            attrs.update(
                name=model_name(model),
            )
        return super(CollectionMeta, mcs).__new__(mcs, name, bases, attrs)


class Collection(APIMixin):

    """DDP Model Collection."""

    __metaclass__ = CollectionMeta

    name = None
    model = None
    qs_filter = None
    order_by = None
    user_rel = None

    def get_queryset(self, base_qs=None):
        """Return a filtered, ordered queryset for this collection."""
        qs = self.model.objects.all() if base_qs is None else base_qs
        # enforce ordering so later use of distinct() works as expected.
        if not qs.query.order_by:
            if self.order_by is None:
                qs = qs.order_by('pk')
            else:
                qs = qs.order_by(*self.order_by)
        if self.qs_filter:
            qs = qs.filter(self.qs_filter)
        return qs

    queryset = property(get_queryset)

    def objects_for_user(self, user, qs=None, xmin__lte=None):
        """Find objects in queryset related to specified user."""
        qs = self.get_queryset(qs)
        user_rels = self.user_rel
        if user_rels:
            if user is None:
                return qs.none()  # no user but we need one: return no objects.
            if isinstance(user_rels, basestring):
                user_rels = [user_rels]
            user_filter = None
            for user_rel in user_rels:
                filter_obj = Q(**{user_rel: user})
                if user_filter is None:
                    user_filter = filter_obj
                else:
                    user_filter |= filter_obj
            qs = qs.filter(user_filter).distinct()
        if xmin__lte is not None:
            qs = qs.extra(
                where=["'xmin' <= %s"],
                params=[xmin__lte],
            )
        return qs

    def user_ids_for_object(self, obj, base_qs=None, include_superusers=True):
        """Find user IDs related to object/pk in queryset."""
        qs = base_qs or self.queryset
        if self.user_rel:
            user_rels = self.user_rel
            if isinstance(user_rels, basestring):
                user_rels = [user_rels]
            user_rel_map = {
                '_user_rel_%d' % index: Array(user_rel)
                for index, user_rel
                in enumerate(user_rels)
            }

            user_ids = set()
            if include_superusers:
                user_ids.update(
                    get_user_model().objects.filter(
                        is_superuser=True, is_active=True,
                    ).values_list('pk', flat=True)
                )

            for rel_user_ids in qs.filter(
                    pk=hasattr(obj, 'pk') and obj.pk or obj,
            ).annotate(**user_rel_map).values_list(*user_rel_map.keys()).get():
                user_ids.update(rel_user_ids)
            user_ids.difference_update([None])
            return user_ids
        else:
            return None

    def field_schema(self):
        """Generate schema for consumption by clients."""
        type_map = {
            'AutoField': 'String',
            'BooleanField': 'Boolean',
            'CharField': 'String',
            'DateTimeField': 'Date',
            'DecimalField': 'Number',
            'FloatField': 'Number',
            'ForeignKey': 'String',
            'PositiveIntegerField': 'Number',
            'TextField': 'String',
        }
        db_type_map = {
            'serial': 'Number',
            'text': 'String',
            'boolean': 'Boolean',
            'integer': 'Number',
        }
        # Django supports model._meta -> pylint: disable=W0212
        meta = self.model._meta
        for field in meta.local_fields:
            int_type = field.get_internal_type()
            schema = {
                'type': (
                    type_map.get(int_type, None)
                ) or (
                    db_type_map.get(field.db_type(connection), 'String')
                )
            }

            rel = getattr(field, 'rel', None)
            if rel:
                schema['type'] = 'String'
                schema['relation'] = {
                    'name': field.name,
                    'collection': model_name(rel.to),
                }

            choices = getattr(field, 'choices', None)
            if choices:
                schema['allowedValues'] = [val for val, _ in choices]
                schema['autoform'] = {
                    'options': [
                        {'label': desc, 'value': val}
                        for val, desc in choices
                    ],
                }

            blank = getattr(field, 'blank', None)
            if blank:
                schema['optional'] = True

            formfield = field.formfield()
            if formfield:
                schema['label'] = force_text(formfield.label)

            max_length = getattr(field, 'max_length', None)
            if max_length is not None:
                schema['max'] = max_length

            if int_type == 'PositiveIntegerField':
                schema['min'] = 0
            if int_type in ('DecimalField', 'FloatField'):
                schema['decimal'] = True
            yield field.column, schema
        for field in meta.local_many_to_many:
            yield '%s_ids' % field.column, {
                'type': '[String]',
                'relation': {
                    'name': field.name,
                    'collection': model_name(field.rel.to),
                },
            }

    @api_endpoint
    def schema(self):
        """Return a representation of the schema for this collection."""
        return {
            name: schema
            for name, schema
            in self.field_schema()
        }

    def serialize(self, obj):
        """Generate a DDP msg for obj with specified msg type."""
        # check for F expressions
        exps = [
            name for name, val in vars(obj).items()
            if isinstance(val, ExpressionNode)
        ]
        if exps:
            # clone/update obj with values but only for the expression fields
            obj = deepcopy(obj)
            for name, val in self.model.objects.values(*exps).get(pk=obj.pk):
                setattr(obj, name, val)

        # run serialization now all fields are "concrete" (not F expressions)
        return this.serializer.serialize([obj])[0]

    def obj_change_as_msg(self, obj, msg):
        """Return DDP change message of specified type (msg) for obj."""
        if msg == 'removed':
            data = {'pk': get_meteor_id(obj)}  # `removed` only needs ID
        elif msg in ('added', 'changed'):
            data = self.serialize(obj)
            data['id'] = str(data.pop('pk'))  # force casting ID as string
        else:
            raise ValueError('Invalid message type: %r' % msg)

        data.update(msg=msg, collection=self.name)
        return data


class PublicationMeta(APIMeta):

    """DDP Publication metaclass."""

    def __new__(mcs, name, bases, attrs):
        """Create a new Publication class."""
        attrs.update(
            api_path_prefix_format='publication/{name}/',
        )
        return super(PublicationMeta, mcs).__new__(mcs, name, bases, attrs)


class Publication(APIMixin):

    """DDP Publication (a set of queries)."""

    __metaclass__ = PublicationMeta

    name = None
    queries = None

    def get_queries(self, *params):
        """DDP get_queries - must override if using params."""
        if params:
            raise NotImplementedError(
                'Publication params not implemented on %r publication.' % (
                    self.name,
                ),
            )
        return self.queries[:]

    @api_endpoint
    def collections(self, *params):
        """Return list of collections for this publication."""
        return sorted(
            set(
                hasattr(qs, 'model') and model_name(qs.model) or qs[1]
                for qs
                in self.get_queries(*params)
            )
        )


def pub_path(publication_name):
    """Return api_path for a publication."""
    return Publication.api_path_prefix_format.format(name=publication_name)


class DDP(APIMixin):

    """Django DDP API."""

    __metaclass__ = APIMeta

    pgworker = None
    _in_migration = False

    class Msg(object):

        """DDP message type enumeration."""

        ADDED = 'added'
        CHANGED = 'changed'
        REMOVED = 'removed'

    def __init__(self):
        """DDP API init."""
        self._registry = {}
        self._subs = {}

    def get_collection(self, model):
        """Return collection instance for given model."""
        name = model_name(model)
        return self.get_col_by_name(name)

    def get_col_by_name(self, name):
        """Return collection instance for given name."""
        return self._registry[COLLECTION_PATH_FORMAT.format(name=name)]

    @property
    def api_providers(self):
        """Return an iterable of API providers."""
        return self._registry.values()

    def sub_notify(self, id_, names, data):
        """Dispatch DDP updates to connections."""
        ws, _ = self._subs[id_]
        connection_pk = data.pop('_sender', None)
        tx_id = data.pop('_tx_id', None)
        connection = getattr(ws, 'connection', None)
        if connection is None:
            tx_id = None
        else:
            if connection.pk != connection_pk:
                tx_id = None
        ws.send_msg(data, tx_id=tx_id)

    def qs_and_collection(self, qs):
        """Return (qs, collection) from qs (which may be a tuple)."""
        if hasattr(qs, 'model'):
            return (qs, self.get_collection(qs.model))
        elif isinstance(qs, (list, tuple)):
            return (qs[0], self.get_col_by_name(qs[1]))
        else:
            raise TypeError('Invalid query spec: %r' % qs)

    @api_endpoint
    def sub(self, id_, name, *params):
        """Create subscription, send matched objects that haven't been sent."""
        try:
            pub = self._registry[pub_path(name)]
        except KeyError:
            this.error('Invalid publication name: %r' % name)
            return
        obj, created = Subscription.objects.get_or_create(
            connection_id=this.ws.connection.pk,
            sub_id=id_,
            user_id=this.request.user.pk,
            defaults={
                'publication': pub.name,
                'params_ejson': ejson.dumps(params),
            },
        )
        if not created:
            this.send_msg({'msg': 'ready', 'subs': [id_]})
            return
        # re-read from DB so we can get transaction ID (xmin)
        obj = Subscription.objects.extra(**XMIN).get(pk=obj.pk)
        queries = collections.OrderedDict(
            (col.name, (col, qs))
            for (qs, col)
            in (
                self.qs_and_collection(qs)
                for qs
                in pub.get_queries(*params)
            )
        )
        self._subs[id_] = (this.ws, sorted(queries))
        self.pgworker.subscribe(self.sub_notify, id_, sorted(queries))
        # mergebox via MVCC!  For details on how this is possible, read this:
        # https://devcenter.heroku.com/articles/postgresql-concurrency
        to_send = collections.OrderedDict(
            (
                name,
                col.objects_for_user(
                    user=this.request.user.pk,
                    qs=qs,
                    xmin__lte=obj.xmin,
                ),
            )
            for name, (col, qs)
            in queries.items()
        )
        for name, (col, qs) in queries.items():
            obj.collections.create(
                model_name=model_name(qs.model),
                collection_name=name,
            )
        for other in Subscription.objects.filter(
                connection=this.ws.connection,
                collections__collection_name__in=queries.keys(),
        ).exclude(
            pk=obj.pk,
        ).order_by('pk').distinct():
            other_pub = self._registry[pub_path(other.publication)]
            for qs in other_pub.get_queries(*other.params):
                qs, col = self.qs_and_collection(qs)
                if col not in to_send:
                    continue
                to_send[col] = to_send[col.name].exclude(
                    pk__in=col.objects_for_user(
                        user=this.request.user.pk,
                        qs=qs,
                        xmin__lte=obj.xmin,
                    ).values('pk'),
                )
        for collection_name, qs in to_send.items():
            col = self.get_col_by_name(collection_name)
            for obj in qs:
                payload = col.obj_change_as_msg(obj, self.Msg.ADDED)
                this.send_msg(payload)
        this.send_msg({'msg': 'ready', 'subs': [id_]})

    def unsub_notify(self, id_):
        """Dispatch DDP updates to connections."""
        (ws, _) = self._subs.pop(id_, (None, []))
        if ws is not None:
            Subscription.objects.filter(
                connection=ws.connection,
                sub_id=id_,
            ).delete()
            ws.send_msg({'msg': 'nosub', 'id': id_})

    @api_endpoint
    def unsub(self, id_):
        """Remove a subscription."""
        self.pgworker.unsubscribe(self.unsub_notify, id_)

    @api_endpoint
    def method(self, method, params, id_):
        """Invoke a method."""
        try:
            handler = self.api_path_map()[method]
        except KeyError:
            this.error('Unknown method: %s' % method)
            return
        try:
            result = handler(*params)
            msg = {'msg': 'result', 'id': id_}
            if result is not None:
                msg['result'] = result
            this.send_msg(msg)
        except Exception, err:  # log error+stack trace -> pylint: disable=W0703
            details = traceback.format_exc()
            this.ws.logger.error(err, exc_info=True)
            msg = {
                'msg': 'result',
                'id': id_,
                'error': {
                    'error': 500,
                    'reason': str(err),
                },
            }
            if settings.DEBUG:
                msg['error']['details'] = details
            this.send_msg(msg)

    def register(self, api_or_iterable):
        """Register an API endpoint."""
        if hasattr(api_or_iterable, 'api_path_prefix'):
            api_or_iterable = [api_or_iterable]
        for api in api_or_iterable:
            api = api()
            if api.api_path_prefix in self._registry:
                raise AlreadyRegistered(
                    'API with prefix %r is already registered to %r' % (
                        api.api_path_prefix,
                        self._registry[api.api_path_prefix],
                    ),
                )
            self._registry[api.api_path_prefix] = api
            self.clear_api_path_map_cache()

    @api_endpoint
    def schema(self):
        """Return schema for all registered collections."""
        res = {}
        for api_provider in self.api_providers:
            if isinstance(api_provider, Collection):
                collection = api_provider
                res[model_name(collection.model)] = collection.schema()
        return res

    def ready(self):
        """Initialisation for django-ddp (setup lookups and signal handlers)."""
        signals.post_save.connect(self.on_save)
        signals.post_delete.connect(self.on_delete)
        signals.m2m_changed.connect(self.on_m2m_changed)
        signals.post_migrate.connect(self.on_post_migrate)

    def on_save(self, sender, **kwargs):
        """Post-save signal handler."""
        if self._in_migration:
            return
        self.send_notify(
            model=sender,
            obj=kwargs['instance'],
            msg=kwargs['created'] and self.Msg.ADDED or self.Msg.CHANGED,
            using=kwargs['using'],
        )

    def on_delete(self, sender, **kwargs):
        """Post-delete signal handler."""
        if self._in_migration:
            return
        self.send_notify(
            model=sender,
            obj=kwargs['instance'],
            msg='removed',
            using=kwargs['using'],
        )

    def on_m2m_changed(self, sender, **kwargs):
        """M2M-changed signal handler."""
        if self._in_migration:
            return
        # See https://docs.djangoproject.com/en/1.7/ref/signals/#m2m-changed
        if kwargs['action'] in (
                'post_add',
                'post_remove',
                'post_clear',
        ):
            if kwargs['reverse'] is False:
                objs = [kwargs['instance']]
                model = objs[0].__class__
            else:
                model = kwargs['model']
                objs = model.objects.filter(pk__in=kwargs['pk_set'])

            for obj in objs:
                self.send_notify(
                    model=model,
                    obj=obj,
                    msg='changed',
                    using=kwargs['using'],
                )

    def on_pre_migrate(self, sender, **kwargs):
        """Pre-migrate signal handler."""
        self._in_migration = True

    def on_post_migrate(self, sender, **kwargs):
        """Post-migrate signal handler."""
        self._in_migration = False
        try:
            Connection.objects.all().delete()
        except DatabaseError:  # pylint: disable=E0712
            pass

    def send_notify(self, model, obj, msg, using):
        """Dispatch PostgreSQL async NOTIFY."""
        col_user_ids = {}
        mod_name = model_name(model)
        if mod_name.split('.', 1)[0] in ('migrations', 'dddp'):
            return  # never send migration or DDP internal models
        col_sub_ids = collections.defaultdict(set)
        for sub in Subscription.objects.filter(
                collections__model_name=mod_name,
        ).prefetch_related('collections'):
            for qs, col in (
                    self.qs_and_collection(qs)
                    for qs
                    in self._registry[
                        'publication/%s/' % sub.publication
                    ].get_queries(*sub.params)
            ):
                # check if obj is an instance of the model for the queryset
                if qs.model is not model:
                    continue  # wrong model on queryset

                # check if obj is included in this subscription
                if not qs.filter(pk=obj.pk).exists():
                    continue  # subscription doesn't include this obj

                # filter qs using user_rel paths on collection
                # retreieve list of allowed users via colleciton
                try:
                    user_ids = col_user_ids[col.__class__]
                except KeyError:
                    user_ids = col_user_ids[col.__class__] = \
                        col.user_ids_for_object(obj)
                if user_ids is None:
                    pass  # unrestricted collection, anyone permitted to see.

                # check if user is in permitted list of users
                if user_ids is not None:
                    pass  # unrestricted collection, anyone permitted to see.
                elif sub.user_id in user_ids:
                    continue  # not for this user

                col_sub_ids[col].add(sub.sub_id)

        if not col_sub_ids:
            get_meteor_id(obj)  # force creation of meteor ID using randomSeed
            return  # no subscribers for this object, nothing more to do.

        for col, sub_ids in col_sub_ids.items():
            payload = col.obj_change_as_msg(obj, msg)
            payload['_sub_ids'] = sorted(sub_ids)
            try:
                ws = this.ws
                payload['_tx_id'] = ws.get_tx_id()
                payload['_sender'] = ws.connection.pk
            except AttributeError:
                pass
            cursor = connections[using].cursor()
            cursor.execute(
                'NOTIFY "%s", %%s' % col.name,
                [
                    ejson.dumps(payload),
                ],
            )


API = DDP()



"""Django DDP PostgreSQL Greenlet."""

from __future__ import absolute_import

import collections

import ejson
import gevent
import gevent.queue
import gevent.select
import psycopg2  # green
import psycopg2.extensions
from geventwebsocket.logging import create_logger


class PostgresGreenlet(gevent.Greenlet):

    """Greenlet for multiplexing database operations."""

    def __init__(self, conn, debug=False):
        """Prepare async connection."""
        super(PostgresGreenlet, self).__init__()
        self.logger = create_logger(__name__, debug=debug)

        # queues for processing incoming sub/unsub requests and processing
        self.subs = gevent.queue.Queue()
        self.unsubs = gevent.queue.Queue()
        self.proc_queue = gevent.queue.Queue()
        self._stop_event = gevent.event.Event()

        # dict of name: subscribers
        # eg: {'bookstore.book': {'tpozNWMPphaJ2n8bj': <function at ...>}}
        self.all_subs = collections.defaultdict(dict)
        self._sub_lock = gevent.lock.RLock()

        # connect to DB in async mode
        conn.allow_thread_sharing = True
        self.connection = conn
        self.conn_params = conn.get_connection_params()
        self.conn_params.update(
            async=True,
        )
        self.conn = psycopg2.connect(**self.conn_params)
        self.poll()  # wait for conneciton to start
        self.cur = self.conn.cursor()

    def _run(self):  # pylint: disable=method-hidden
        """Spawn sub tasks, wait for stop signal."""
        gevent.spawn(self.process_conn)
        gevent.spawn(self.process_subs)
        gevent.spawn(self.process_unsubs)
        self._stop_event.wait()

    def stop(self):
        """Stop subtasks and let run() finish."""
        self._stop_event.set()

    def subscribe(self, func, id_, names):
        """Register callback `func` to be called after NOTIFY for `names`."""
        self.subs.put((func, id_, names))

    def unsubscribe(self, func, id_):
        """Un-register callback `func` to be called after NOTIFY for `id`."""
        self.unsubs.put((func, id_))

    def process_conn(self):
        """Subtask to process NOTIFY async events from DB connection."""
        while not self._stop_event.is_set():
            # TODO: change timeout so self._stop_event is periodically checked?
            gevent.select.select([self.conn], [], [], timeout=None)
            self.poll()

    def poll(self):
        """Poll DB socket and process async tasks."""
        while 1:
            state = self.conn.poll()
            if state == psycopg2.extensions.POLL_OK:
                while self.conn.notifies:
                    notify = self.conn.notifies.pop()
                    name = notify.channel
                    self.logger.info(
                        "Got NOTIFY (pid=%d, name=%r, payload=%r)",
                        notify.pid, name, notify.payload,
                    )
                    try:
                        self._sub_lock.acquire()
                        self.logger.info(self.all_subs)
                        subs = self.all_subs[name]
                        data = ejson.loads(notify.payload)
                        sub_ids = data.pop('_sub_ids')
                        self.logger.info('Subscribers: %r', sub_ids)
                        self.logger.info(subs)
                        for id_, func in subs.items():
                            if id_ not in sub_ids:
                                continue  # not for this subscription
                            gevent.spawn(func, id_, name, data)
                    finally:
                        self._sub_lock.release()
                break
            elif state == psycopg2.extensions.POLL_WRITE:
                gevent.select.select([], [self.conn.fileno()], [])
            elif state == psycopg2.extensions.POLL_READ:
                gevent.select.select([self.conn.fileno()], [], [])
            else:
                self.logger.warn('POLL_ERR: %s', state)

    def process_subs(self):
        """Subtask to process `sub` requests from `self.subs` queue."""
        while not self._stop_event.is_set():
            func, id_, names = self.subs.get()
            try:
                self._sub_lock.acquire()
                for name in names:
                    subs = self.all_subs[name]
                    if len(subs) == 0:
                        self.logger.debug('LISTEN "%s";', name)
                        self.poll()
                        self.cur.execute('LISTEN "%s";' % name)
                        self.poll()
                    subs[id_] = func
            finally:
                self._sub_lock.release()

    def process_unsubs(self):
        """Subtask to process `unsub` requests from `self.unsubs` queue."""
        while not self._stop_event.is_set():
            func, id_ = self.unsubs.get()
            try:
                self._sub_lock.acquire()
                for name in list(self.all_subs):
                    subs = self.all_subs[name]
                    subs.pop(id_, None)
                    if len(subs) == 0:
                        self.logger.info('UNLISTEN "%s";', name)
                        self.cur.execute('UNLISTEN "%s";' % name)
                        self.poll()
                        del self.all_subs[name]
            finally:
                self._sub_lock.release()
            gevent.spawn(func, id_)



"""Django DDP WebSocket service."""

from __future__ import print_function, absolute_import

import collections
import sys


Addr = collections.namedtuple('Addr', ['host', 'port'])


def ddpp_sockjs_xhr(environ, start_response):
    """Dummy method that doesn't handle XHR requests."""
    start_response(
        '404 Not found',
        [
            ('Content-Type', 'text/plain; charset=UTF-8'),
            (
                'Access-Control-Allow-Origin',
                '/'.join(environ['HTTP_REFERER'].split('/')[:3]),
            ),
            ('Access-Control-Allow-Credentials', 'true'),
            #  ('access-control-allow-credentials', 'true'),
            ('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0'),
            ('Connection', 'keep-alive'),
            ('Vary', 'Origin'),
        ],
    )
    yield 'No.'


def ddpp_sockjs_info(environ, start_response):
    """Inform client that WebSocket service is available."""
    import random
    import ejson

    start_response(
        '200 OK',
        [
            ('Content-Type', 'application/json; charset=UTF-8'),
            (
                'Access-Control-Allow-Origin',
                '/'.join(environ['HTTP_REFERER'].split('/')[:3]),
            ),
            ('Access-Control-Allow-Credentials', 'true'),
            #  ('access-control-allow-credentials', 'true'),
            ('Cache-Control', 'no-store, no-cache, must-revalidate, max-age=0'),
            ('Connection', 'keep-alive'),
            ('Vary', 'Origin'),
        ],
    )
    yield ejson.dumps(collections.OrderedDict([
        ('websocket', True),
        ('origins', [
            '*:*',
        ]),
        ('cookie_needed', False),
        ('entropy', random.getrandbits(32)),
    ]))


def serve(listen, debug=False):
    """Spawn greenlets for handling websockets and PostgreSQL calls."""
    import signal
    from django.apps import apps
    from django.db import connection, close_old_connections
    from django.utils.module_loading import import_string
    from dddp.postgres import PostgresGreenlet
    from dddp.websocket import DDPWebSocketApplication
    import gevent
    import geventwebsocket

    # shutdown existing connections
    close_old_connections()

    # setup PostgresGreenlet to multiplex DB calls
    pgworker = PostgresGreenlet(connection, debug=debug)
    DDPWebSocketApplication.pgworker = pgworker

    # use settings.WSGI_APPLICATION or fallback to default Django WSGI app
    from django.conf import settings
    if hasattr(settings, 'WSGI_APPLICATION'):
        wsgi_name = settings.WSGI_APPLICATION
        wsgi_app = import_string(wsgi_name)
    else:
        from django.core.wsgi import get_wsgi_application
        wsgi_app = get_wsgi_application()
        wsgi_name = str(wsgi_app.__class__)

    resource = geventwebsocket.Resource({
        r'/websocket': DDPWebSocketApplication,
        r'^/sockjs/\d+/\w+/websocket$': DDPWebSocketApplication,
        r'^/sockjs/\d+/\w+/xhr$': ddpp_sockjs_xhr,
        r'^/sockjs/info$': ddpp_sockjs_info,
        r'^/(?!(websocket|sockjs)/)': wsgi_app,
    })

    # setup WebSocketServer to dispatch web requests
    webservers = [
        geventwebsocket.WebSocketServer(
            (host, port),
            resource,
            debug=debug,
        )
        for host, port
        in listen
    ]

    def killall(*args, **kwargs):
        """Kill all green threads."""
        pgworker.stop()
        for webserver in webservers:
            webserver.stop()

    # die gracefully with SIGINT or SIGQUIT
    gevent.signal(signal.SIGINT, killall)
    gevent.signal(signal.SIGQUIT, killall)

    print('=> Discovering DDP endpoints...')
    api = apps.get_app_config('dddp').api
    api.pgworker = pgworker
    DDPWebSocketApplication.api = api
    print(
        '\n'.join(
            '    %s' % api_path
            for api_path
            in sorted(api.api_path_map())
        ),
    )

    # start greenlets
    pgworker.start()
    print('=> Started PostgresGreenlet.')
    web_threads = [
        gevent.spawn(webserver.serve_forever)
        for webserver
        in webservers
    ]
    print('=> Started DDPWebSocketApplication.')
    print('=> Started your app (%s).' % wsgi_name)
    print('')
    for host, port in listen:
        print('=> App running at: http://%s:%d/' % (host, port))
    gevent.joinall(web_threads)
    pgworker.stop()
    gevent.joinall([pgworker])


def addr(val, default_port=8000, defualt_host='localhost'):
    """
    Convert a string of format host[:port] into Addr(host, port).

    >>> addr('0:80')
    Addr(host='0', port=80)

    >>> addr('127.0.0.1:80')
    Addr(host='127.0.0.1', port=80)

    >>> addr('0.0.0.0', default_port=8000)
    Addr(host='0.0.0.0', port=8000)
    """
    import re
    import socket

    match = re.match(r'\A(?P<host>.*?)(:(?P<port>(\d+|\w+)))?\Z', val)
    if match is None:
        raise argparse.ArgumentTypeError(
            '%r is not a valid host[:port] address.' % val
        )
    host, port = match.group('host', 'port')

    if not host:
        host = defualt_host

    if not port:
        port = default_port
    elif port.isdigit():
        port = int(port)
    else:
        port = socket.getservbyname(port)

    return Addr(host, port)


def main():
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        'listen', metavar='address[:port]', nargs='*', type=addr,
    )
    namespace = parser.parse_args()
    serve(namespace.listen or [Addr('localhost', 8000)])


if __name__ == '__main__':
    from dddp import greenify
    greenify()
    main()



#!/usr/bin/env python
"""
Alea PRNG.

This implementation of Alea defaults to a more secure initial internal state.
>>> r1, r2 = Alea(), Alea()
>>> assert r1.state != r2.state, 'r1: %r, r2: %r' % (r1.state, r2.state)

>>> random = Alea("my", 3, "seeds")
>>> (random.s0, random.s1, random.s2)
(0.23922116006724536, 0.6147655111271888, 0.3493568613193929)

>>> random()
0.30802189325913787

>>> random()
0.5190450621303171

>>> random()
0.43635262292809784


>>> random = Alea("my", 3, "seeds")
>>> random()
0.30802189325913787

>>> random = Alea("my", 3, "seeds")

>>> random.random_string(17, UNMISTAKABLE)
'JYRduBwQtjpeCkqP7'

>>> random.random_string(17, UNMISTAKABLE)
'HLxYtpZBtSain84zj'

>>> random.random_string(17, UNMISTAKABLE)
's9XrbWaDC4yCL5NCW'

>>> random.random_string(17, UNMISTAKABLE)
'SCiymgNnZpwda9vSH'

>>> random.random_string(17, UNMISTAKABLE)
'hui3ThSoZrFrdFDTT'


>>> random = Alea("my", 3, "seeds")

>>> random.random_string(43, BASE64)
'tHBM5k8z4TZOmU0zgsv9H4ZIl4CJSXic_T3iF2KFJnm'

"""

from math import floor
import os
import random
import time

UNMISTAKABLE = '23456789ABCDEFGHJKLMNPQRSTWXYZabcdefghijkmnopqrstuvwxyz'
BASE64 = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_'


class Mash(object):

    """
    `Mash` hasing algorithm.

    >>> mash = Mash()
    >>> mash(' ')
    0.8633289230056107
    >>> mash(' ')
    0.15019597788341343
    >>> mash(' ')
    0.9176952994894236
    """

    def __init__(self):
        """Initialise state."""
        self.n = 0xefc8249d

    def __call__(self, data):
        """Return mash, updating internal state."""
        data = bytes(data)
        for byte in bytes(data):
            self.n += ord(byte)
            h = 0.02519603282416938 * self.n
            self.n = floor(h)
            h -= self.n
            h *= self.n
            self.n = floor(h)
            h -= self.n
            self.n += h * 0x100000000
        res = self.n * 2.3283064365386963e-10  # 2^-32
        return res


class Alea(object):

    """Alea stateful PRNG."""

    c = None
    s0 = None
    s1 = None
    s2 = None

    def __init__(self, *args):
        """Initialise Alea state from seeds (args)."""
        self.seed(args)

    def seed(self, values):
        """Seed internal state from supplied values."""
        if not values:
            # Meteor uses epoch seconds as the seed if no args supplied, we use
            # a much more secure seed by default to avoid hash collisions.
            seed_ids = [int, str, random, self, values, self.__class__]
            random.shuffle(seed_ids)
            values = map(id, seed_ids) + [time.time(), os.urandom(512)]

        mash = Mash()
        self.c = 1
        self.s0 = mash(' ')
        self.s1 = mash(' ')
        self.s2 = mash(' ')

        for val in values:
            self.s0 -= mash(val)
            if self.s0 < 0:
                self.s0 += 1
            self.s1 -= mash(val)
            if self.s1 < 0:
                self.s1 += 1
            self.s2 -= mash(val)
            if self.s2 < 0:
                self.s2 += 1

    @property
    def state(self):
        """Return internal state, useful for testing."""
        return {'c': self.c, 's0': self.s0, 's1': self.s1, 's2': self.s2}

    def __call__(self):
        """Get the next psuedo random number, updating state."""
        t = 2091639 * self.s0 + self.c * 2.3283064365386963e-10  # 2^-32
        self.c = floor(t)
        self.s0 = self.s1
        self.s1 = self.s2
        self.s2 = t - self.c
        return self.s2

    def choice(self, seq):
        """Choose an element from the sequence `seq`."""
        return seq[int(self() * len(seq))]

    def random_string(self, length, alphabet):
        """Return string of `length` elements chosen from `alphabet`."""
        return ''.join(
            self.choice(alphabet) for n in range(length)
        )

    def hex_string(self, digits):
        """Return a hex string of `digits` length."""
        return self.random_string(digits, '0123456789abcdef')


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""Django DDP app config."""

from __future__ import print_function

from django.apps import AppConfig
from django.core import serializers
from django.conf import settings, ImproperlyConfigured
from django.db import DatabaseError
from django.db.models import signals

from dddp import autodiscover
from dddp.models import Connection


class DjangoDDPConfig(AppConfig):

    """Django app config for django-ddp."""

    api = None
    name = 'dddp'
    verbose_name = 'Django DDP'
    _in_migration = False

    def ready(self):
        """Initialisation for django-ddp (setup lookups and signal handlers)."""
        serializers.register_serializer('ddp', 'dddp.serializer')
        if not settings.DATABASES:
            raise ImproperlyConfigured('No databases configured.')
        for (alias, conf) in settings.DATABASES.items():
            if conf['ENGINE'] != 'django.db.backends.postgresql_psycopg2':
                raise ImproperlyConfigured(
                    '%r uses %r: django-ddp only works with PostgreSQL.' % (
                        alias, conf['backend'],
                    )
                )
        self.api = autodiscover()
        self.api.ready()



#!/usr/bin/env python
"""Django/PostgreSQL implementation of the Meteor DDP service."""
import os.path
from setuptools import setup, find_packages

setup(
    name='django-ddp',
    version='0.6.0',
    description=__doc__,
    long_description=open('README.rst').read(),
    author='Tyson Clugg',
    author_email='tyson@clugg.net',
    url='https://github.com/commoncode/django-ddp',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Django>=1.7,<1.8',
        'psycopg2>=2.5.4',
        'gevent>=1.0',
        'gevent-websocket>=0.9',
        'meteor-ejson>=1.0',
        'psycogreen>=1.0',
        'django-dbarray>=0.2',
    ],
    entry_points={
        'console_scripts': [
            'dddp=dddp.main:main',
        ],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries",
    ],
)



"""Django DDP WebSocket service."""

from __future__ import absolute_import

import atexit
import inspect
import itertools
import sys
import traceback

from six.moves import range as irange

import ejson
import geventwebsocket
from django.core.handlers.base import BaseHandler
from django.core.handlers.wsgi import WSGIRequest
from django.db import connection, transaction

from dddp import THREAD_LOCAL as this, alea


class MeteorError(Exception):

    """MeteorError."""

    pass


def validate_kwargs(func, kwargs):
    """Validate arguments to be supplied to func."""
    func_name = func.__name__
    argspec = inspect.getargspec(func)
    all_args = argspec.args[:]
    defaults = list(argspec.defaults or [])

    # ignore implicit 'self' argument
    if inspect.ismethod(func) and all_args[:1] == ['self']:
        all_args[:1] = []

    # don't require arguments that have defaults
    if defaults:
        required = all_args[:-len(defaults)]
    else:
        required = all_args[:]

    # translate 'foo_' to avoid reserved names like 'id'
    trans = {
        arg: arg.endswith('_') and arg[:-1] or arg
        for arg
        in all_args
    }
    for key in list(kwargs):
        key_adj = '%s_' % key
        if key_adj in all_args:
            kwargs[key_adj] = kwargs.pop(key)

    # figure out what we're missing
    supplied = sorted(kwargs)
    missing = [
        trans.get(arg, arg) for arg in required
        if arg not in supplied
    ]
    if missing:
        raise MeteorError(
            'Missing required arguments to %s: %s' % (
                func_name,
                ' '.join(missing),
            ),
            getattr(func, 'err', None),
        )

    # figure out what is extra
    extra = [
        arg for arg in supplied
        if arg not in all_args
    ]
    if extra:
        raise MeteorError(
            'Unknown arguments to %s: %s' % (
                func_name,
                ' '.join(extra),
            ),
        )


class DDPWebSocketApplication(geventwebsocket.WebSocketApplication):

    """Django DDP WebSocket application."""

    _tx_buffer = None
    _tx_buffer_id_gen = None
    _tx_next_id_gen = None
    _tx_next_id = None

    methods = {}
    versions = [  # first item is preferred version
        '1',
        'pre1',
        'pre2',
    ]
    api = None
    logger = None
    pgworker = None
    remote_addr = None
    version = None
    support = None
    connection = None
    subs = None
    request = None
    base_handler = BaseHandler()

    def get_tx_id(self):
        """Get the next TX msg ID."""
        return next(self._tx_buffer_id_gen)

    def on_open(self):
        """Handle new websocket connection."""
        self.logger = self.ws.logger

        # self._tx_buffer collects outgoing messages which must be sent in order
        self._tx_buffer = {}
        # track the head of the queue (buffer) and the next msg to be sent
        self._tx_buffer_id_gen = itertools.cycle(irange(sys.maxint))
        self._tx_next_id_gen = itertools.cycle(irange(sys.maxint))
        # start by waiting for the very first message
        self._tx_next_id = next(self._tx_next_id_gen)

        self.request = WSGIRequest(self.ws.environ)
        # Apply request middleware (so we get request.user and other attrs)
        # pylint: disable=protected-access
        if self.base_handler._request_middleware is None:
            self.base_handler.load_middleware()
        for middleware_method in self.base_handler._request_middleware:
            response = middleware_method(self.request)
            if response:
                raise ValueError(response)
        this.ws = self
        this.request = self.request
        this.send = self.send
        this.send_msg = self.send_msg
        this.reply = self.reply
        this.error = self.error
        this.request.session.save()

        this.remote_addr = self.remote_addr = \
            '{0[REMOTE_ADDR]}:{0[REMOTE_PORT]}'.format(
                self.ws.environ,
            )
        self.subs = {}
        self.logger.info('+ %s OPEN %s', self, this.request.user)
        self.send('o')
        self.send('a["{\\"server_id\\":\\"0\\"}"]')

    def __str__(self):
        """Show remote address that connected to us."""
        return self.remote_addr

    def on_close(self, reason):
        """Handle closing of websocket connection."""
        if self.connection is not None:
            self.connection.delete()
            self.connection = None
        self.logger.info('- %s %s', self, reason or 'CLOSE')

    def on_message(self, message):
        """Process a message received from remote."""
        if self.ws.closed:
            return None
        try:
            self.logger.debug('< %s %r', self, message)

            # parse message set
            try:
                msgs = ejson.loads(message)
            except ValueError, err:
                raise MeteorError(400, 'Data is not valid EJSON')
            if not isinstance(msgs, list):
                raise MeteorError(400, 'Invalid EJSON messages')

            # process individual messages
            while msgs:
                # parse message payload
                raw = msgs.pop(0)
                try:
                    try:
                        data = ejson.loads(raw)
                    except ValueError, err:
                        raise MeteorError(400, 'Data is not valid EJSON')
                    if not isinstance(data, dict):
                        self.error(400, 'Invalid EJSON message payload', raw)
                        continue
                    try:
                        msg = data.pop('msg')
                    except KeyError:
                        raise MeteorError(
                            400, 'Bad request', None, {'offendingMessage': data}
                        )
                    # dispatch message
                    self.dispatch(msg, data)
                except MeteorError, err:
                    traceback.print_exc()
                    self.error(err)
                except Exception, err:
                    traceback.print_exc()
                    self.error(err)
        except geventwebsocket.WebSocketError, err:
            self.ws.close()
        except MeteorError, err:
            self.error(err)

    @transaction.atomic
    def dispatch(self, msg, kwargs):
        """Dispatch msg to appropriate recv_foo handler."""
        # enforce calling 'connect' first
        if self.connection is None and msg != 'connect':
            raise MeteorError(400, 'Must connect first')

        # lookup method handler
        try:
            handler = getattr(self, 'recv_%s' % msg)
        except (AttributeError, UnicodeEncodeError):
            raise MeteorError(404, 'Method not found')

        # validate handler arguments
        validate_kwargs(handler, kwargs)

        # dispatch to handler
        try:
            handler(**kwargs)
        except Exception, err:  # print stack trace --> pylint: disable=W0703
            traceback.print_exc()
            self.error(MeteorError(500, 'Internal server error', err))

    def send(self, data, tx_id=None):
        """Send raw `data` to WebSocket client."""
        # buffer data until we get pre-requisite data
        if tx_id is None:
            tx_id = self.get_tx_id()
        self._tx_buffer[tx_id] = data

        # de-queue messages from buffer
        while self._tx_next_id in self._tx_buffer:
            # pull next message from buffer
            data = self._tx_buffer.pop(self._tx_next_id)
            # advance next message ID
            self._tx_next_id = next(self._tx_next_id_gen)
            # send message
            self.logger.debug('> %s %r', self, data)
            try:
                self.ws.send(data)
            except geventwebsocket.WebSocketError:
                self.ws.close()
                break

    def send_msg(self, payload, tx_id=None):
        """Send EJSON payload to remote."""
        data = ejson.dumps([ejson.dumps(payload)])
        self.send('a%s' % data, tx_id=tx_id)

    def reply(self, msg, **kwargs):
        """Send EJSON reply to remote."""
        kwargs['msg'] = msg
        self.send_msg(kwargs)

    def error(self, err, reason=None, detail=None, **kwargs):
        """Send EJSON error to remote."""
        if isinstance(err, MeteorError):
            (
                err, reason, detail, kwargs,
            ) = (
                err.args[:] + (None, None, None, None)
            )[:4]
        elif isinstance(err, Exception):
            reason = str(err)
        data = {
            'error': '%s' % (err or '<UNKNOWN_ERROR>'),
        }
        if reason:
            if reason is Exception:
                reason = str(reason)
            data['reason'] = reason
        if detail:
            if isinstance(detail, Exception):
                detail = str(detail)
            data['detail'] = detail
        if kwargs:
            data.update(kwargs)
        self.logger.error('! %s %r', self, data)
        self.reply('error', **data)

    def recv_connect(self, version=None, support=None, session=None):
        """DDP connect handler."""
        if self.connection is not None:
            self.error(
                'Session already established.',
                reason='Current session in detail.',
                detail=self.connection.connection_id,
            )
        elif None in (version, support) or version not in self.versions:
            self.reply('failed', version=self.versions[0])
        elif version not in support:
            self.error('Client version/support mismatch.')
        else:
            from dddp.models import Connection
            cur = connection.cursor()
            cur.execute('SELECT pg_backend_pid()')
            (backend_pid,) = cur.fetchone()
            this.version = version
            this.support = support
            self.connection = Connection.objects.create(
                server_addr='%d:%s' % (
                    backend_pid,
                    self.ws.handler.socket.getsockname(),
                ),
                remote_addr=self.remote_addr,
                version=version,
            )
            atexit.register(self.on_close, 'Shutting down.')
            self.reply('connected', session=self.connection.connection_id)

    def recv_ping(self, id_=None):
        """DDP ping handler."""
        if id_ is None:
            self.reply('pong')
        else:
            self.reply('pong', id=id_)

    def recv_sub(self, id_, name, params):
        """DDP sub handler."""
        self.api.sub(id_, name, *params)
    recv_sub.err = 'Malformed subscription'

    def recv_unsub(self, id_=None):
        """DDP unsub handler."""
        if id_:
            self.api.unsub(id_)
        else:
            self.reply('nosub')

    def recv_method(self, method, params, id_, randomSeed=None):
        """DDP method handler."""
        if randomSeed is not None:
            this.random_streams.random_seed = randomSeed
            this.alea_random = alea.Alea(randomSeed)
        self.api.method(method, params, id_)
        self.reply('updated', methods=[id_])
    recv_method.err = 'Malformed method invocation'
